/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "mex.h"
#include "mp4ff.h"
#include "common.h"
#include "structs.h"
#include "decoder.h"
#include "dct.h"
#include "filters.h"

uint32_t read_callback(void *user_data, void *buffer, uint32_t length)
{
    return fread(buffer, 1, length, (FILE*)user_data);
}

uint32_t seek_callback(void *user_data, uint64_t position)
{
    return fseek((FILE*)user_data, position, SEEK_SET);
}

void mfcc_compute( double *in, double *out, int num_coefs ) {

	int m,l;
	double amp;
	double eps = 1e-16;
	double earmag[40];

	/* Filterbank */
	for ( m=0 ; m<40 ; m++ ) {
		earmag[m] = 0.0;
		for ( l=filters_first[m] ; l<=filters_last[m] ; l++ ) {
			amp = in[l];
			if (amp<0) amp = - amp;
			earmag[m] += filters_weight[m*1024+l] * amp;
			
		}
		earmag[m] = log10( earmag[m] + eps );
	}

	/* DCT */
	for ( m=0 ; m<num_coefs ; m++ ) {
		out[m] = 0.0;
		for ( l=0 ; l<40 ; l++ ) {
			out[m] += dct[m*40+l] * earmag[l];
		}
	}
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {

	/**********************************/
	/********** Declarations **********/
	/**********************************/

	int k,l;
	char *filename;
	int filename_len;
	FILE *mp4file;
	mp4ff_t *infile;
	int n,numframes;
	unsigned char *buffer;
	unsigned int buffer_size;
	NeAACDecHandle hDecoder;
	NeAACDecFrameInfo info;
	unsigned int samplerate;
	unsigned char channels;
	long buffercount;
	void *sample_buffer;
	double *coefs = NULL;
	double *winseq = NULL;
	double *winshape = NULL;
	double input[1024];
	double output[40];
	int rc;
	long dur;
	int numcoefs;
	double scale = pow(2,-20.5);
	int channel;

	/**********************************/
	/************* Init ***************/
	/**********************************/

	/* Check number of arguments */
	if (nrhs<1) mexErrMsgTxt("Not enough input arguments.");
	if (nlhs>3) mexErrMsgTxt("Too many output arguments.");

	/* Open input file */
	filename_len = mxGetN(prhs[0])+1;
	filename = (char*)mxCalloc(filename_len, sizeof(char));
	mxGetString(prhs[0],filename,filename_len);
	mp4file = fopen(filename,"rb");
	if (mp4file==NULL)
		mexErrMsgTxt("Can't open input file\n");

	/* Coefs number */
	numcoefs = 13;
	if (nrhs>=2) numcoefs = (int)mxGetPr(prhs[1])[0];
	if (numcoefs<1||numcoefs>40) mexErrMsgTxt("Number of coefficients must be in the range [1,40].\n");

	/* Channel */
	channel = 0;
	if (nrhs>=3) channel = (int)mxGetPr(prhs[2])[0];
	if (channel>1||channel<0) mexErrMsgTxt( "Channel parameter must be 0 or 1.");

	/* Init MPEG-4 decoder */
	mp4ff_callback_t *mp4cb = mxCalloc(1,sizeof(mp4ff_callback_t));
	mp4cb->read = read_callback;
	mp4cb->seek = seek_callback;
	mp4cb->user_data = mp4file;
	infile = mp4ff_open_read(mp4cb);
	if (!infile)
		mexErrMsgTxt("Can't open input file\n");
	if (mp4ff_total_tracks(infile)!=1)
		mexErrMsgTxt("Invalid file\n");

	/* Init AAC decoder */
	buffer = NULL;
	buffer_size = 0;
	mp4ff_get_decoder_config(infile, 0, &buffer, &buffer_size);
	hDecoder = NeAACDecOpen();
	hDecoder->config.outputFormat  = FAAD_FMT_DOUBLE;
	buffercount = NeAACDecInit2(hDecoder, buffer, buffer_size, &samplerate, &channels);
	if (buffer) free(buffer);
	if (channels!=2)
		mexErrMsgTxt("Invalid file\n");

	/* Init output */
	numframes = mp4ff_num_samples(infile,0);
	if (nlhs>0) {
		plhs[0] = mxCreateDoubleMatrix(numcoefs,numframes,mxREAL);
		coefs = mxGetPr(plhs[0]);
	}
	if (nlhs>1) {
		plhs[1] = mxCreateDoubleMatrix(1,numframes,mxREAL);
		winseq = mxGetPr(plhs[1]);
	}
	if (nlhs>2) {
		plhs[2] = mxCreateDoubleMatrix(1,numframes,mxREAL);
		winshape = mxGetPr(plhs[2]);
	}

	/**********************************/
	/********** Decoding **************/
	/**********************************/

	for (n = 0; n < numframes; n++)	{
		/* Read MPEG-4 frame */
		buffer = NULL;
		buffer_size = 0;
		dur = mp4ff_get_sample_duration(infile, 0, n);
		rc = mp4ff_read_sample(infile, 0, n, &buffer,  &buffer_size);
		/* Decode AAC */
		sample_buffer = NeAACDecDecode(hDecoder, &info, buffer, buffer_size, 8);
		if (buffer) free(buffer);
		if (info.object_type!=2) mexErrMsgTxt("Invalid MPEG-4 object type\n");
		/* MFCC */
		if (hDecoder->fb->window_sequence[channel]==2) {
			for (l=0 ; l<8 ; l++)
				for (k=0 ; k<128 ; k++)
					input[k*8+l] = (double)hDecoder->fb->spec_coef[channel][l*128+k]*sqrt(8)*scale;
		} else {
			for (k=0;k<1024;k++) input[k] = (double)hDecoder->fb->spec_coef[channel][k]*scale;
		}
		mfcc_compute( input, output, numcoefs );
		/* Output */
		if (nlhs>0)
			for (k=0;k<numcoefs;k++) coefs[n*numcoefs+k] = output[k];
		if (nlhs>1)
			winseq[n] = hDecoder->fb->window_sequence[channel];
		if (nlhs>2)
			winshape[n] = hDecoder->fb->window_shape[channel];
	}

	/**********************************/
	/*********** Clean up *************/
	/**********************************/

	NeAACDecClose(hDecoder);
	mp4ff_close(infile);
	mxFree(mp4cb);
	fclose(mp4file);
	mxFree(filename);
}

