/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "mex.h"
#include "mp4ff.h"
#include "common.h"
#include "structs.h"
#include "decoder.h"

uint32_t read_callback(void *user_data, void *buffer, uint32_t length)
{
    return fread(buffer, 1, length, (FILE*)user_data);
}

uint32_t seek_callback(void *user_data, uint64_t position)
{
    return fseek((FILE*)user_data, position, SEEK_SET);
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {

	/**********************************/
	/********** Declarations **********/
	/**********************************/

	int k,numsamples;
	char *filename;
	int filename_len;
	FILE *mp4file;
	mp4ff_t *infile;
	int n,numframes;
	unsigned char *buffer;
	unsigned int buffer_size;
	NeAACDecHandle hDecoder;
	NeAACDecFrameInfo info;
	unsigned int samplerate;
	unsigned char channels;
	long buffercount;
	void *sample_buffer;
	double *sample_buffer_double;
	double *signal = NULL;
	int rc;
	long dur;
	int outputlen = 0;
	int numchannels = 0;
	int delay;

	/**********************************/
	/************* Init ***************/
	/**********************************/

	/* Check number of arguments */
	if (nrhs<1) mexErrMsgTxt( "Not enough input arguments.");
	if (nlhs>2) mexErrMsgTxt( "Too many output arguments.");

	/* Input file */
	filename_len = mxGetN(prhs[0])+1;
	filename = (char*)mxCalloc(filename_len, sizeof(char));
	mxGetString(prhs[0],filename,filename_len);
	mp4file = fopen(filename,"rb");
	if (mp4file==NULL)
		mexErrMsgTxt("Can't open input file\n");

	/* Delay */
	delay = 0;
	if (nrhs>1) delay = (int)mxGetPr(prhs[1])[0];
	if (delay<0)
		mexErrMsgTxt("Invalid delay\n");

	/* Init MPEG-4 decoder */
	mp4ff_callback_t *mp4cb = mxCalloc(1,sizeof(mp4ff_callback_t));
	mp4cb->read = read_callback;
	mp4cb->seek = seek_callback;
	mp4cb->user_data = mp4file;
	infile = mp4ff_open_read(mp4cb);
	if (!infile)
		mexErrMsgTxt("Can't open input file\n");
	if (mp4ff_total_tracks(infile)!=1)
		mexErrMsgTxt("Invalid file\n");

	/* Init AAC decoder */
	buffer = NULL;
	buffer_size = 0;
	mp4ff_get_decoder_config(infile, 0, &buffer, &buffer_size);
	hDecoder = NeAACDecOpen();
	hDecoder->config.outputFormat  = FAAD_FMT_DOUBLE;
	buffercount = NeAACDecInit2(hDecoder, buffer, buffer_size, &samplerate, &channels);
	if (buffer) free(buffer);
	if (channels!=2)
		mexErrMsgTxt("Invalid file\n");

	/* Init output */
	numframes = mp4ff_num_samples(infile,0);
	numsamples = 0;

	/**********************************/
	/********** Decoding **************/
	/**********************************/

	for (n = 0; n < numframes; n++)	{
		/* Read MPEG-4 frame */
		buffer = NULL;
		buffer_size = 0;
		dur = mp4ff_get_sample_duration(infile, 0, n);
		rc = mp4ff_read_sample(infile, 0, n, &buffer,  &buffer_size);
		/* Decode AAC */
		sample_buffer = NeAACDecDecode(hDecoder, &info, buffer, buffer_size, 0);
		if (buffer) free(buffer);
		if (info.object_type!=2) mexErrMsgTxt("Invalid MPEG-4 object type\n");
		/* Output samples */
		if (nlhs>=1&&info.samples > 0) {
			if (numsamples==0) {
				numchannels = hDecoder->fr_channels;
				outputlen = numframes*1024 - delay;
				plhs[0] = mxCreateDoubleMatrix(outputlen,numchannels,mxREAL);
				signal = mxGetPr(plhs[0]);
			}
			sample_buffer_double = (double*)sample_buffer;
			for (k=0;k<1024;k++) {
				if (numsamples>=delay) {
					if (numchannels==1) {
						signal[numsamples-delay] = sample_buffer_double[2*k];
					} else {
						signal[(numsamples-delay)] = sample_buffer_double[2*k];
						signal[outputlen+(numsamples-delay)] = sample_buffer_double[2*k+1];
					}
				}
				numsamples++;
			}
		}
	}

	/* Output sample rate */
	if (nlhs>=2) {
		plhs[1] = mxCreateDoubleMatrix(1,1,mxREAL);
		mxGetPr(plhs[1])[0] = (double)samplerate;
	}

	/**********************************/
	/*********** Clean up *************/
	/**********************************/

	NeAACDecClose(hDecoder);
	mp4ff_close(infile);
	mxFree(mp4cb);
	fclose(mp4file);
	mxFree(filename);
}

